@@ -18,6 +18,10 @@
  * Copyright (C) 2005 - 2008 Red Hat, Inc.
  */
 
+#define _XOPEN_SOURCE
+#include <time.h>
+#undef _XOPEN_SOURCE
+
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <dbus/dbus.h>
@@ -28,6 +32,7 @@
 #include <stdio.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#include <ctype.h>
 
 #include <config.h>
 
@@ -40,6 +45,8 @@
 
 #if defined(TARGET_DEBIAN)
 #define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhcp3"
+#elif defined(TARGET_SUSE) || defined(TARGET_MANDRIVA)
+#define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhcp"
 #else
 #define NM_DHCP_MANAGER_LEASE_DIR       LOCALSTATEDIR "/lib/dhclient"
 #endif
@@ -61,19 +68,221 @@ get_pidfile_for_iface (const char * iface)
 
 
 static char *
-get_leasefile_for_iface (const char * iface)
+get_leasefile_for_iface (const char * iface, const char *uuid)
 {
-	return g_strdup_printf ("%s/%s-%s.%s",
+	return g_strdup_printf ("%s/%s-%s-%s.%s",
 	                        NM_DHCP_MANAGER_LEASE_DIR,
 	                        NM_DHCP_MANAGER_LEASE_FILENAME,
+	                        uuid,
 	                        iface,
 	                        NM_DHCP_MANAGER_LEASE_FILE_EXT);
 }
 
+static void
+add_lease_option (GHashTable *hash, char *line)
+{
+	char *spc;
+
+	spc = strchr (line, ' ');
+	if (!spc) {
+		g_warning ("%s: line '%s' did not contain a space", __func__, line);
+		return;
+	}
+
+	/* If it's an 'option' line, split at second space */
+	if (g_str_has_prefix (line, "option ")) {
+		spc = strchr (spc + 1, ' ');
+		if (!spc) {
+			g_warning ("%s: option line '%s' did not contain a second space",
+			           __func__, line);
+			return;
+		}
+	}
+
+	/* Split the line at the space */
+	*spc = '\0';
+	spc++;
+
+	/* Kill the ';' at the end of the line, if any */
+	if (*(spc + strlen (spc) - 1) == ';')
+		*(spc + strlen (spc) - 1) = '\0';
+
+	/* Treat 'interface' specially */
+	if (g_str_has_prefix (line, "interface")) {
+		if (*(spc) == '"')
+			spc++; /* Jump past the " */
+		if (*(spc + strlen (spc) - 1) == '"')
+			*(spc + strlen (spc) - 1) = '\0';  /* Kill trailing " */
+	}
+
+	g_hash_table_insert (hash, g_strdup (line), g_strdup (spc));
+}
+
+GSList *
+nm_dhcp_client_get_lease_ip4_config (const char *iface, const char *uuid)
+{
+	GSList *parsed = NULL, *iter, *leases = NULL;
+	char *contents = NULL;
+	char *leasefile;
+	char **line, **split = NULL;
+	GHashTable *hash = NULL;
+
+	leasefile = get_leasefile_for_iface (iface, uuid);
+	if (!leasefile)
+		return NULL;
+
+	if (!g_file_test (leasefile, G_FILE_TEST_EXISTS))
+		goto out;
+
+	if (!g_file_get_contents (leasefile, &contents, NULL, NULL))
+		goto out;
+
+	split = g_strsplit_set (contents, "\n\r", -1);
+	g_free (contents);
+	if (!split)
+		goto out;
+
+	for (line = split; line && *line; line++) {
+		*line = g_strstrip (*line);
+
+		if (!strcmp (*line, "}")) {
+			/* Lease ends */
+			parsed = g_slist_append (parsed, hash);
+			hash = NULL;
+		} else if (!strcmp (*line, "lease {")) {
+			/* Beginning of a new lease */
+			if (hash) {
+				g_warning ("%s: lease file %s malformed; new lease started "
+				           "without ending previous lease",
+				           __func__, leasefile);
+				g_hash_table_destroy (hash);
+			}
+
+			hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+		} else if (strlen (*line))
+			add_lease_option (hash, *line);
+	}
+	g_strfreev (split);
+
+	/* Check if the last lease in the file was properly ended */
+	if (hash) {
+		g_warning ("%s: lease file %s malformed; new lease started "
+		           "without ending previous lease",
+		           __func__, leasefile);
+		g_hash_table_destroy (hash);
+		hash = NULL;
+	}
+
+	for (iter = parsed; iter; iter = g_slist_next (iter)) {
+		NMIP4Config *ip4;
+		NMIP4Address *addr;
+		const char *data;
+		struct in_addr tmp;
+		guint32 prefix;
+		struct tm expire;
+
+		hash = iter->data;
+
+		/* Make sure this lease is for the interface we want */
+		data = g_hash_table_lookup (hash, "interface");
+		if (!data || strcmp (data, iface))
+			continue;
+
+		data = g_hash_table_lookup (hash, "expire");
+		if (data) {
+			time_t now_tt;
+			struct tm *now;
+
+			/* Read lease expiration (in UTC) */
+			if (!strptime (data, "%w %Y/%m/%d %H:%M:%S", &expire)) {
+				g_warning ("%s: couldn't parse expire time '%s'",
+				           __func__, data);
+				continue;
+			}
+
+			now_tt = time (NULL);
+			now = gmtime(&now_tt);
+
+			/* Ignore this lease if it's already expired */
+			if (expire.tm_year < now->tm_year)
+				continue;
+			else if (expire.tm_year == now->tm_year) {
+				if (expire.tm_mon < now->tm_mon)
+					continue;
+				else if (expire.tm_mon == now->tm_mon) {
+					if (expire.tm_mday < now->tm_mday)
+						continue;
+					else if (expire.tm_mday == now->tm_mday) {
+						if (expire.tm_hour < now->tm_hour)
+							continue;
+						else if (expire.tm_hour == now->tm_hour) {
+							if (expire.tm_min < now->tm_min)
+								continue;
+							else if (expire.tm_min == now->tm_min) {
+								if (expire.tm_sec <= now->tm_sec)
+									continue;
+							}
+						}
+					}
+				}
+			}
+			/* If we get this far, the lease hasn't expired */
+		}
+
+		data = g_hash_table_lookup (hash, "fixed-address");
+		if (!data)
+			continue;
+
+		ip4 = nm_ip4_config_new ();
+		addr = nm_ip4_address_new ();
+
+		/* IP4 address */
+		if (!inet_pton (AF_INET, data, &tmp)) {
+			g_warning ("%s: couldn't parse IP4 address '%s'", __func__, data);
+			goto error;
+		}
+		nm_ip4_address_set_address (addr, tmp.s_addr);
+
+		/* Netmask */
+		data = g_hash_table_lookup (hash, "option subnet-mask");
+		if (!data)
+			data = "255.255.255.0"; /* FIXME: assume class C? */
+		if (!inet_pton (AF_INET, data, &tmp)) {
+			g_warning ("%s: couldn't parse IP4 subnet mask '%s'", __func__, data);
+			goto error;
+		}
+		prefix = nm_utils_ip4_netmask_to_prefix (tmp.s_addr);
+		nm_ip4_address_set_prefix (addr, prefix);
+
+		/* Gateway */
+		data = g_hash_table_lookup (hash, "option routers");
+		if (data) {
+			if (!inet_pton (AF_INET, data, &tmp)) {
+				g_warning ("%s: couldn't parse IP4 gateway '%s'", __func__, data);
+				goto error;
+			}
+			nm_ip4_address_set_gateway (addr, tmp.s_addr);
+		}
+
+		nm_ip4_config_take_address (ip4, addr);
+		leases = g_slist_append (leases, ip4);
+		continue;
+
+	error:
+		nm_ip4_address_unref (addr);
+		g_object_unref (ip4);
+	}
+
+out:
+	g_slist_foreach (parsed, (GFunc) g_hash_table_destroy, NULL);
+	g_free (leasefile);
+	return leases;
+}
 
 
 #define DHCP_CLIENT_ID_TAG "send dhcp-client-identifier"
 #define DHCP_CLIENT_ID_FORMAT DHCP_CLIENT_ID_TAG " \"%s\"; # added by NetworkManager"
+#define DHCP_CLIENT_ID_FORMAT_OCTETS DHCP_CLIENT_ID_TAG " %s; # added by NetworkManager"
 
 #define DHCP_HOSTNAME_TAG "send host-name"
 #define DHCP_HOSTNAME_FORMAT DHCP_HOSTNAME_TAG " \"%s\"; # added by NetworkManager"
@@ -81,6 +290,7 @@ get_leasefile_for_iface (const char * iface)
 static gboolean
 merge_dhclient_config (NMDHCPDevice *device,
                        NMSettingIP4Config *s_ip4,
+                       guint8 *anycast_addr,
                        const char *contents,
                        const char *orig,
                        GError **error)
@@ -132,14 +342,44 @@ merge_dhclient_config (NMDHCPDevice *device,
 		const char *tmp;
 
 		tmp = nm_setting_ip4_config_get_dhcp_client_id (s_ip4);
-		if (tmp)
-			g_string_append_printf (new_contents, DHCP_CLIENT_ID_FORMAT "\n", tmp);
+		if (tmp) {
+			gboolean is_octets = TRUE;
+			const char *p = tmp;
+
+			while (*p) {
+				if (!isxdigit (*p) && (*p != ':')) {
+					is_octets = FALSE;
+					break;
+				}
+				p++;
+			}
+
+			/* If the client ID is just hex digits and : then don't use quotes,
+			 * becuase dhclient expects either a quoted ASCII string, or a byte
+			 * array formated as hex octets separated by :
+			 */
+			if (is_octets)
+				g_string_append_printf (new_contents, DHCP_CLIENT_ID_FORMAT_OCTETS "\n", tmp);
+			else
+				g_string_append_printf (new_contents, DHCP_CLIENT_ID_FORMAT "\n", tmp);
+		}
 
 		tmp = nm_setting_ip4_config_get_dhcp_hostname (s_ip4);
 		if (tmp)
 			g_string_append_printf (new_contents, DHCP_HOSTNAME_FORMAT "\n", tmp);
 	}
 
+	if (anycast_addr) {
+		g_string_append_printf (new_contents, "interface \"%s\" {\n"
+		                        " initial-interval 1; \n"
+		                        " anycast-mac ethernet %02x:%02x:%02x:%02x:%02x:%02x;\n"
+		                        "}\n",
+		                        device->iface,
+		                        anycast_addr[0], anycast_addr[1],
+		                        anycast_addr[2], anycast_addr[3],
+		                        anycast_addr[4], anycast_addr[5]);
+	}
+
 	if (g_file_set_contents (device->conf_file, new_contents->str, -1, error))
 		success = TRUE;
 
@@ -154,7 +394,9 @@ merge_dhclient_config (NMDHCPDevice *device,
  * config file along with the NM options.
  */
 static gboolean
-create_dhclient_config (NMDHCPDevice *device, NMSettingIP4Config *s_ip4)
+create_dhclient_config (NMDHCPDevice *device,
+                        NMSettingIP4Config *s_ip4,
+                        guint8 *dhcp_anycast_addr)
 {
 	char *orig = NULL, *contents = NULL;
 	GError *error = NULL;
@@ -167,6 +409,8 @@ create_dhclient_config (NMDHCPDevice *device, NMSettingIP4Config *s_ip4)
 	orig = g_strdup (SYSCONFDIR "/dhclient.conf");
 #elif defined(TARGET_DEBIAN)
 	orig = g_strdup (SYSCONFDIR "/dhcp3/dhclient.conf");
+#elif defined(TARGET_GENTOO)
+	orig = g_strdup (SYSCONFDIR "/dhcp/dhclient.conf");
 #else
 	orig = g_strdup_printf (SYSCONFDIR "/dhclient-%s.conf", device->iface);
 #endif
@@ -192,7 +436,7 @@ create_dhclient_config (NMDHCPDevice *device, NMSettingIP4Config *s_ip4)
 
 out:
 	error = NULL;
-	if (merge_dhclient_config (device, s_ip4, contents, orig, &error))
+	if (merge_dhclient_config (device, s_ip4, dhcp_anycast_addr, contents, orig, &error))
 		success = TRUE;
 	else {
 		nm_warning ("%s: error creating dhclient configuration: %s",
@@ -216,7 +460,10 @@ dhclient_child_setup (gpointer user_data G_GNUC_UNUSED)
 
 
 GPid
-nm_dhcp_client_start (NMDHCPDevice *device, NMSettingIP4Config *s_ip4)
+nm_dhcp_client_start (NMDHCPDevice *device,
+                      const char *uuid,
+                      NMSettingIP4Config *s_ip4,
+                      guint8 *dhcp_anycast_addr)
 {
 	GPtrArray *dhclient_argv = NULL;
 	GPid pid = 0;
@@ -234,13 +481,13 @@ nm_dhcp_client_start (NMDHCPDevice *device, NMSettingIP4Config *s_ip4)
 		goto out;
 	}
 
-	device->lease_file = get_leasefile_for_iface (device->iface);
+	device->lease_file = get_leasefile_for_iface (device->iface, uuid);
 	if (!device->lease_file) {
 		nm_warning ("%s: not enough memory for dhclient options.", device->iface);
 		goto out;
 	}
 
-	if (!create_dhclient_config (device, s_ip4))
+	if (!create_dhclient_config (device, s_ip4, dhcp_anycast_addr))
 		goto out;
 
 	/* Kill any existing dhclient bound to this interface */
