@@ -40,6 +40,7 @@
 #include "nm-marshal.h"
 #include "nm-utils.h"
 #include "nm-dbus-manager.h"
+#include "nm-hostname-provider.h"
 #include "nm-dbus-glib-types.h"
 #include "nm-glib-compat.h"
 
@@ -52,6 +53,7 @@ typedef struct {
 	NMDBusManager * dbus_mgr;
 	GHashTable *	devices;
 	DBusGProxy *	proxy;
+	NMHostnameProvider *hostname_provider;
 } NMDHCPManagerPrivate;
 
 
@@ -71,6 +73,8 @@ static NMDHCPManager *nm_dhcp_manager_new (void);
 
 static void nm_dhcp_manager_cancel_transaction_real (NMDHCPDevice *device);
 
+static void hostname_provider_destroyed (gpointer data, GObject *destroyed_object);
+
 NMDHCPManager *
 nm_dhcp_manager_get (void)
 {
@@ -95,6 +99,11 @@ finalize (GObject *object)
 {
 	NMDHCPManagerPrivate *priv = NM_DHCP_MANAGER_GET_PRIVATE (object);
 
+	if (priv->hostname_provider) {
+		g_object_weak_unref (G_OBJECT (priv->hostname_provider), hostname_provider_destroyed, object);
+		priv->hostname_provider = NULL;
+	}
+
 	g_hash_table_destroy (priv->devices);
 	g_object_unref (priv->proxy);
 	g_object_unref (priv->dbus_mgr);
@@ -169,7 +178,7 @@ nm_dhcp_device_watch_cleanup (NMDHCPDevice * device)
 static void
 nm_dhcp_device_destroy (NMDHCPDevice *device)
 {
-	int ret;
+	int ignored;
 
 	nm_dhcp_device_timeout_cleanup (device);
 
@@ -180,7 +189,7 @@ nm_dhcp_device_destroy (NMDHCPDevice *device)
 		g_hash_table_destroy (device->options);
 
 	if (device->conf_file) {
-		ret = unlink (device->conf_file);
+		ignored = unlink (device->conf_file);
 		g_free (device->conf_file);
 	}
 
@@ -583,6 +592,7 @@ nm_dhcp_manager_begin_transaction (NMDHCPManager *manager,
 {
 	NMDHCPManagerPrivate *priv;
 	NMDHCPDevice *device;
+	NMSettingIP4Config *setting;
 
 	g_return_val_if_fail (NM_IS_DHCP_MANAGER (manager), FALSE);
 	g_return_val_if_fail (iface != NULL, FALSE);
@@ -598,13 +608,33 @@ nm_dhcp_manager_begin_transaction (NMDHCPManager *manager,
 		nm_dhcp_manager_cancel_transaction_real (device);
 	}
 
+	if (s_ip4 && 
+		nm_setting_ip4_config_get_dhcp_send_hostname (s_ip4) &&
+		nm_setting_ip4_config_get_dhcp_hostname (s_ip4) == NULL &&
+		priv->hostname_provider != NULL) {
+		/* We're asked to send the hostname to DHCP server,
+		   the hostname isn't specified,
+		   and a hostname provider is registered: use that */
+
+		setting = NM_SETTING_IP4_CONFIG (nm_setting_duplicate (NM_SETTING (s_ip4)));
+		g_object_set (G_OBJECT (setting),
+					  NM_SETTING_IP4_CONFIG_DHCP_HOSTNAME,
+					  nm_hostname_provider_get_hostname (priv->hostname_provider),
+					  NULL);
+	} else {
+		setting = s_ip4 ? g_object_ref (s_ip4) : NULL;
+	}
+
 	if (timeout == 0)
 		timeout = NM_DHCP_TIMEOUT;
 
 	nm_info ("Activation (%s) Beginning DHCP transaction (timeout in %d seconds)",
 	         iface, timeout);
+	device->pid = nm_dhcp_client_start (device, uuid, setting, dhcp_anycast_addr);
+
+	if (setting)
+		g_object_unref (setting);
 
-	device->pid = nm_dhcp_client_start (device, uuid, s_ip4, dhcp_anycast_addr);
 	if (device->pid == 0)
 		return FALSE;
 
@@ -669,7 +699,7 @@ nm_dhcp_manager_cancel_transaction_real (NMDHCPDevice *device)
 
 	nm_dhcp_client_stop (device, device->pid);
 
-	nm_info ("%s: canceled DHCP transaction, dhcp client pid %d",
+	nm_info ("(%s): canceled DHCP transaction, dhcp client pid %d",
 	         device->iface,
 	         device->pid);
 
@@ -819,7 +849,7 @@ nm_dhcp_manager_options_to_ip4_config (const char *iface, GHashTable *options)
 	struct in_addr tmp_addr;
 	NMIP4Address *addr = NULL;
 	char *str = NULL;
-	guint32 gwaddr = 0, prefix = 0;
+	guint32 gwaddr = 0;
 	gboolean have_classless = FALSE;
 
 	g_return_val_if_fail (iface != NULL, NULL);
@@ -846,14 +876,9 @@ nm_dhcp_manager_options_to_ip4_config (const char *iface, GHashTable *options)
 
 	str = g_hash_table_lookup (options, "new_subnet_mask");
 	if (str && (inet_pton (AF_INET, str, &tmp_addr) > 0)) {
-		prefix = nm_utils_ip4_netmask_to_prefix (tmp_addr.s_addr);
-		nm_info ("  prefix %d (%s)", prefix, str);
-	} else {
-		/* Get default netmask for the IP according to appropriate class. */
-		prefix = nm_utils_ip4_get_default_prefix (nm_ip4_address_get_address (addr));
-		nm_info ("  prefix %d (default)", prefix);
+		nm_ip4_address_set_prefix (addr, nm_utils_ip4_netmask_to_prefix (tmp_addr.s_addr));
+		nm_info ("  prefix %d (%s)", nm_ip4_address_get_prefix (addr), str);
 	}
-	nm_ip4_address_set_prefix (addr, prefix);
 
 	/* Routes: if the server returns classless static routes, we MUST ignore
 	 * the 'static_routes' option.
@@ -1070,3 +1095,42 @@ nm_dhcp_manager_foreach_dhcp4_option (NMDHCPManager *self,
 	return TRUE;
 }
 
+static void
+hostname_provider_destroyed (gpointer data, GObject *destroyed_object)
+{
+	NM_DHCP_MANAGER_GET_PRIVATE (data)->hostname_provider = NULL;
+}
+
+void
+nm_dhcp_manager_set_hostname_provider (NMDHCPManager *manager,
+									   NMHostnameProvider *provider)
+{
+	NMDHCPManagerPrivate *priv;
+
+	g_return_if_fail (NM_IS_DHCP_MANAGER (manager));
+
+	priv = NM_DHCP_MANAGER_GET_PRIVATE (manager);
+
+	if (priv->hostname_provider) {
+		g_object_weak_unref (G_OBJECT (priv->hostname_provider), hostname_provider_destroyed, manager);
+		priv->hostname_provider = NULL;
+	}
+
+	if (provider) {
+		priv->hostname_provider = provider;
+		g_object_weak_ref (G_OBJECT (provider), hostname_provider_destroyed, manager);
+	}
+}
+
+GSList *
+nm_dhcp_manager_get_lease_ip4_config (NMDHCPManager *self,
+                                      const char *iface,
+                                      const char *uuid)
+{
+	g_return_val_if_fail (NM_IS_DHCP_MANAGER (self), NULL);
+	g_return_val_if_fail (iface != NULL, NULL);
+	g_return_val_if_fail (uuid != NULL, NULL);
+
+	return nm_dhcp_client_get_lease_ip4_config (iface, uuid);
+}
+
